use crate::grpc::challenge::NodeUpdate;
use ed25519_dalek::{Signer, SigningKey, VerifyingKey};
use rand::rngs::OsRng;
use std::collections::{HashMap, HashSet};
use std::time::Duration;
use std::time::SystemTime;
use std::time::UNIX_EPOCH;
use tabled::{Table, Tabled};
use tokio::sync::Mutex;

#[derive(Clone, PartialEq, Debug)]
pub struct NodeInfo {
    pub pubkey: VerifyingKey,
    pub updated_at: SystemTime,
    pub public: bool,
}

/// Needs to be surrounded in an Arc.
pub struct Store {
    nodes: Mutex<HashMap<IP, NodeInfo>>,
    conns: Mutex<HashSet<IP>>,
    keys: Mutex<HashMap<VerifyingKey, SigningKey>>,
}
pub enum SigningError {
    CorruptedKey,
    KeyNotFound,
}

impl From<hex::FromHexError> for SigningError {
    fn from(_: hex::FromHexError) -> Self {
        Self::CorruptedKey
    }
}

impl From<ed25519_dalek::ed25519::Error> for SigningError {
    fn from(_: ed25519_dalek::ed25519::Error) -> Self {
        Self::CorruptedKey
    }
}

impl From<std::array::TryFromSliceError> for SigningError {
    fn from(_: std::array::TryFromSliceError) -> Self {
        Self::CorruptedKey
    }
}

type IP = String;

impl std::fmt::Display for SigningError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let error_message = match self {
            SigningError::CorruptedKey => "The public key is corrupted",
            SigningError::KeyNotFound => "Did not find the public key",
        };
        write!(f, "{}", error_message)
    }
}

impl Store {
    // app should exit if any error happens here so unwrap() is good
    pub fn init() -> Self {
        Self {
            nodes: Mutex::new(HashMap::new()),
            keys: Mutex::new(HashMap::new()),
            conns: Mutex::new(HashSet::new()),
        }
    }

    pub async fn add_conn(&self, ip: &str) {
        let mut conns = self.conns.lock().await;
        conns.insert(ip.to_string());
    }

    pub async fn delete_conn(&self, ip: &str) {
        let mut conns = self.conns.lock().await;
        conns.remove(ip);
    }

    pub async fn tabled_node_list(&self) -> String {
        #[derive(Tabled)]
        struct OutputRow {
            ip: String,
            pubkey: String,
            age: u64,
            public: bool,
        }
        let mut output = vec![];
        for (ip, node_info) in self.nodes.lock().await.iter() {
            let ip = ip.clone();
            let pubkey = hex::encode(node_info.pubkey.as_bytes());
            let age = std::time::SystemTime::now()
                .duration_since(node_info.updated_at)
                .unwrap_or(Duration::ZERO)
                .as_secs();
            let public = node_info.public;
            output.push(OutputRow {
                ip,
                pubkey,
                age,
                public,
            });
        }
        Table::new(output).to_string()
    }

    pub async fn sign_message_with_key(
        &self,
        message: &str,
        pubkey: &str,
    ) -> Result<String, SigningError> {
        let key_bytes = hex::decode(pubkey)?;
        let pubkey = VerifyingKey::from_bytes(&key_bytes.as_slice().try_into()?)?;

        let signing_key = match self.get_privkey(&pubkey).await {
            Some(k) => k,
            None => return Err(SigningError::KeyNotFound),
        };

        // TODO: check if to_bytes returns the signature in a format that people can verify from bash
        let signature = hex::encode(signing_key.sign(message.as_bytes()).to_bytes());

        Ok(signature)
    }

    pub async fn add_key(&self, pubkey: VerifyingKey, privkey: SigningKey) {
        let mut keys = self.keys.lock().await;
        keys.insert(pubkey, privkey);
    }

    pub async fn remove_key(&self, pubkey: &VerifyingKey) {
        let mut keys = self.keys.lock().await;
        keys.remove(pubkey);
    }

    async fn get_privkey(&self, pubkey: &VerifyingKey) -> Option<SigningKey> {
        let keys = self.keys.lock().await;
        keys.get(pubkey).cloned()
    }

    /// This returns true if NodeInfo got modified.
    ///
    /// On a side note, there are two types of people in this worlds:
    ///     1. Those that can extrapolate...
    pub async fn process_grpc_update(&self, node: NodeUpdate) -> bool {
        let key_bytes = match hex::decode(node.keypair.clone()) {
            Ok(k) => k,
            Err(_) => return false,
        };
        let privkey = SigningKey::from_bytes(match &key_bytes.as_slice().try_into() {
            Ok(p) => p,
            Err(_) => return false,
        });
        let pubkey = privkey.verifying_key();
        let updated_at_std: std::time::SystemTime = match node.updated_at {
            Some(ts) => {
                let duration = Duration::new(ts.seconds as u64, ts.nanos as u32);
                UNIX_EPOCH
                    .checked_add(duration)
                    .unwrap_or(SystemTime::now())
            }
            None => SystemTime::now(),
        };

        self.add_key(pubkey, privkey).await;
        let node_info = NodeInfo {
            pubkey,
            updated_at: updated_at_std,
            public: node.public,
        };
        if let Some(mut old_node_info) = self.update_node(node.ip, node_info.clone()).await {
            if !node_info.public {
                old_node_info.public = false;
            }
            match old_node_info.ne(&node_info) {
                true => {
                    self.remove_key(&old_node_info.pubkey).await;
                    true
                }
                false => false,
            }
        } else {
            true
        }
    }

    /// returns old pubkey if node got updated
    async fn update_node(&self, ip: String, info: NodeInfo) -> Option<NodeInfo> {
        let mut nodes = self.nodes.lock().await;
        nodes.insert(ip, info.clone())
    }

    //    pub async fn remove_node(&self, ip: &str) {
    //        let mut nodes = self.nodes.lock().await;
    //        nodes.remove(ip);
    //    }

    pub async fn get_localhost(&self) -> NodeUpdate {
        // these unwrap never fail
        // you could trigger reset_localhost_keys on error, though
        // TODO: do that so that code looks clean
        let nodes = self.nodes.lock().await;
        let keys = self.keys.lock().await;
        let node_info = nodes.get("localhost").unwrap();
        let keypair = keys.get(&node_info.pubkey).unwrap();
        NodeUpdate {
            ip: "localhost".to_string(),
            keypair: hex::encode(keypair.as_bytes()),
            updated_at: Some(prost_types::Timestamp::from(node_info.updated_at)),
            public: false,
        }
    }

    /// freshes the keys of the node and returns a protobuf for the network
    pub async fn reset_localhost_keys(&self) -> NodeUpdate {
        let mut csprng = OsRng;
        let keypair_raw = ed25519_dalek::SigningKey::generate(&mut csprng);
        let keypair = hex::encode(keypair_raw.as_bytes());
        let pubkey = keypair_raw.verifying_key();
        let ip = "localhost".to_string();
        let updated_at = std::time::SystemTime::now();
        let public = false;
        self.update_node(
            ip.clone(),
            NodeInfo {
                pubkey,
                updated_at,
                public,
            },
        )
        .await;
        self.add_key(pubkey, keypair_raw.clone()).await;
        let updated_at = Some(prost_types::Timestamp::from(updated_at));
        NodeUpdate {
            ip,
            keypair,
            updated_at,
            public,
        }
    }

    pub async fn get_full_node_list(&self) -> Vec<NodeUpdate> {
        let nodes = self.nodes.lock().await;
        let keys = self.keys.lock().await;
        nodes
            .iter()
            .filter_map(|(ip, node_info)| {
                keys.get(&node_info.pubkey).map(|signing_key| NodeUpdate {
                    ip: ip.to_string(),
                    keypair: hex::encode(signing_key.as_bytes()),
                    updated_at: Some(prost_types::Timestamp::from(node_info.updated_at)),
                    public: node_info.public,
                })
            })
            .collect()
    }

    // returns a random node that does not have an active connection
    pub async fn get_random_node(&self) -> Option<String> {
        use rand::rngs::OsRng;
        use rand::RngCore;
        let nodes = self.nodes.lock().await;
        let conns = self.conns.lock().await;
        let len = nodes.len();
        if len == 0 {
            return None;
        }
        let skip = OsRng.next_u64().try_into().unwrap_or(0) % len;
        nodes
            .keys()
            .cycle()
            .skip(skip)
            .filter(|k| !conns.contains(*k))
            .next()
            .cloned()
    }
}
